#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#define SIZE	80

char *website = "www.metawire.org";
char *destdir = NULL;

char *current_version (void)
{
	static char buff[SIZE];
	char *cmd = "/usr/local/sbin/recv";
	FILE *fp = NULL;
	struct stat s;
	
	memset(buff, 0, SIZE);
	sprintf(buff, "%s -V", cmd);

	
	if ((stat(cmd, &s))<0) {
		perror("stat");
		exit(0);
	}
	
	if ((fp = popen (buff, "r")) == NULL) {
		perror("popen");
		exit(0);
	}
	
	fgets(buff, SIZE, fp);	
	buff[strlen(buff)-1] = '\0';
	return buff;
}

static char *web_version (void)
{
	struct hostent *server_host = NULL;
	struct sockaddr_in server_addr;
	int server, ret;
	char *s = "GET /~iptgraph/version.php\n";
	static char buff[SIZE];

	if ((server = socket(AF_INET, SOCK_STREAM, 0))<0) {
		perror("socket");
		exit(0);
	}

	if ((server_host = gethostbyname(website))==NULL) {
		perror("gethostbyname");
		exit(0);
	}

	memcpy ((void *) &server_addr.sin_addr, 
			(void *) server_host->h_addr,
			server_host->h_length);

	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons(SIZE);

	if ((ret = connect(server, 
				(struct sockaddr *) &server_addr, 
				sizeof(server_addr)))<0) {
		perror("connect");
		exit (0);
	}

	memset (buff, 0, SIZE);
	send (server, s, strlen(s), MSG_DONTWAIT);
	recv (server, &buff, SIZE, 0);
	
	return buff;	
}

int check_for_updates (void)
{
	int ret = 0;
	char *cv = current_version();
	char *wv = web_version();

	ret = strncmp(cv, wv, strlen(cv));

	// wierd, because current version is more than web's version
	if (ret > 0) {
		return 0;
	// need to update
	} else if (ret < 0) {
		return 1;
	} else {
		return 0;
	}
}

void clearfile (void)
{
	char buff[SIZE];

	if (!destdir) 
		return;
	
	sprintf(buff, "%s/UPDATE", destdir);
	unlink(buff);	
}

void updatefile (char *s)
{
	char buff[SIZE];
	FILE *fp = NULL;

	if (!destdir) 
		return;

	sprintf(buff, "%s/UPDATE", destdir);
	if ((fp = fopen(buff, "w+"))==NULL) {
		perror("fopen");
		exit(0);
	}

	fprintf(fp, "%s", s);
	fclose(fp);
}

void help (void)
{
	printf("checkupdate -d <directory>\n");
	printf("options:\n");
	printf("-d <directory>		: dest directory\n");
	exit(0);
}

int verify_directory (char *dir)
{
	struct stat buf;

	if (dir == NULL)
		help();

	if (lstat(dir, &buf)<0) {
		perror("lstat");
		exit(0);
	}

	if (!S_ISDIR(buf.st_mode)) 
		return 0;
	else 
		return 1;
}

int main (int argc, char **argv)
{
	int c;
	int len;

	destdir = NULL;

	while ((c = getopt(argc, argv, "d:")) != EOF) {
		switch (c) {
			case 'd':
				destdir = optarg;
				len = strlen(destdir)-1;
				break;
			default:
				help();
				break;
		}
	}

	/* remove trailing slashes */
	while (*(destdir+len) == '/') {
		*(destdir+len) = '\0';
		len--;
	}

	if (!verify_directory(destdir)) 
		help();

	clearfile();
	if (check_for_updates()) {
		updatefile(web_version());
	}
	return 0;
}
