#include "graph.h"

char *myhostname (void)
{
	static char buff[80];
	gethostname (buff, 80);
	return buff;
}

void print_stats (struct total_stats *t)
{
	printf("ip:   drop: %-8ld (%-8ld bytes) accept: %-8ld (%-8ld bytes)\n", 
					DROP_COUNT(ip), DROP_BYTE(ip), ACCEPT_COUNT(ip), ACCEPT_BYTE(ip));
	printf("icmp: drop: %-8ld (%-8ld bytes) accept: %-8ld (%-8ld bytes)\n", 
					DROP_COUNT(icmp), DROP_BYTE(icmp), ACCEPT_COUNT(icmp), ACCEPT_BYTE(icmp));
	printf("tcp:  drop: %-8ld (%-8ld bytes) accept: %-8ld (%-8ld bytes)\n", 
					DROP_COUNT(tcp), DROP_BYTE(tcp), ACCEPT_COUNT(tcp), ACCEPT_BYTE(tcp));
	printf("udp:  drop: %-8ld (%-8ld bytes) accept: %-8ld (%-8ld bytes)\n", 
					DROP_COUNT(udp), DROP_BYTE(udp), ACCEPT_COUNT(udp), ACCEPT_BYTE(udp));
}

void create_rrd (void)
{
	struct stat statbuf;
	char buff[1024];
	unsigned long tt = time (NULL);

	if (stat("tcp.rrd", &statbuf)<0) {
		memset (buff, 0, 1024);
		sprintf (buff, "/usr/bin/rrdtool create tcp.rrd --start %ld --step %d DS:tcp_accept:GAUGE:%d:U:U DS:tcp_drop:GAUGE:%d:U:U ", tt, STEP, STEP, STEP);
		strcat (buff, "RRA:LAST:0.5:1:1000 RRA:AVERAGE:0.5:1:1000"); 
		system (buff);
	}

	if (stat("udp.rrd", &statbuf)<0) {
		memset (buff, 0, 1024);
		sprintf (buff, "/usr/bin/rrdtool create udp.rrd --start %ld --step %d DS:udp_accept:GAUGE:%d:U:U DS:udp_drop:GAUGE:%d:U:U ", tt, STEP, STEP, STEP);
		strcat (buff, "RRA:LAST:0.5:1:1000 RRA:AVERAGE:0.5:1:1000"); 
		system (buff);
	}

	if (stat("icmp.rrd", &statbuf)<0) {
		memset (buff, 0, 1024);
		sprintf (buff, "/usr/bin/rrdtool create icmp.rrd --start %ld --step %d DS:icmp_accept:GAUGE:%d:U:U DS:icmp_drop:GAUGE:%d:U:U ", tt, STEP, STEP, STEP);
		strcat (buff, "RRA:LAST:0.5:1:1000 RRA:AVERAGE:0.5:1:1000"); 
		system (buff);
	}

	if (stat("ip.rrd", &statbuf)<0) {
		memset (buff, 0, 1024);
		sprintf (buff, "/usr/bin/rrdtool create ip.rrd --start %ld --step %d DS:ip_accept:GAUGE:%d:U:U DS:ip_drop:GAUGE:%d:U:U ", tt, STEP, STEP, STEP);
		strcat (buff, "RRA:LAST:0.5:1:1000 RRA:AVERAGE:0.5:1:1000"); 
		system (buff);
	}

}

void update_rrd_stats (struct total_stats *t)
{
	char buff[1024];
	unsigned long tt = time (NULL);

	// tcp
	memset (buff, 0, 1024);
	sprintf(buff, "/usr/bin/rrdtool update tcp.rrd %ld:%ld:%ld", tt, ACCEPT_BYTE(tcp), DROP_BYTE(tcp));
	system (buff);
	memset (buff, 0, 1024);
	sprintf(buff, "/usr/bin/rrdtool graph tcp.png --start %ld --end %ld --title=\"TCP Activity (%s)\" --vertical-label \"Bytes\" ", 
					tt-GRAPH_LEN, tt, hostname);
	strcat (buff, "DEF:ta=tcp.rrd:tcp_accept:LAST DEF:td=tcp.rrd:tcp_drop:LAST \"CDEF:test=ta,50,+,50,-\" ");
	strcat (buff, "AREA:test#CCDDFF LINE2:ta#3300FF:\"TCP Accept\" ");
	strcat (buff, "'GPRINT:ta:LAST:\%6.0lf bytes' AREA:td#FF0033:\"TCP Drop\" 'GPRINT:td:LAST:\%6.0lf bytes\\j' ");
	strcat (buff, "> /dev/null");
	system (buff);

	// udp
	memset (buff, 0, 1024);
	sprintf(buff, "/usr/bin/rrdtool update udp.rrd %ld:%ld:%ld", tt, ACCEPT_BYTE(udp), DROP_BYTE(udp));
	system (buff);
	memset (buff, 0, 1024);
	sprintf(buff, "/usr/bin/rrdtool graph udp.png --start %ld --end %ld --title=\"UDP Activity (%s)\" --vertical-label \"Bytes\" ",
				   tt-GRAPH_LEN, tt, hostname);
	strcat (buff, "DEF:ta=udp.rrd:udp_accept:LAST DEF:td=udp.rrd:udp_drop:LAST \"CDEF:test=ta,50,+,50,-\" ");
	strcat (buff, "AREA:test#CCDDFF LINE2:ta#3300FF:\"UDP Accept\" ");
	strcat (buff, "'GPRINT:ta:LAST:\%6.0lf bytes' AREA:td#FF0033:\"UDP Drop\" 'GPRINT:td:LAST:\%6.0lf bytes\\j' ");
	strcat (buff, "> /dev/null");
	system (buff);

	// icmp
	memset (buff, 0, 1024);
	sprintf(buff, "/usr/bin/rrdtool update icmp.rrd %ld:%ld:%ld", tt, ACCEPT_BYTE(icmp), DROP_BYTE(icmp));
	system (buff);
	memset (buff, 0, 1024);
	sprintf(buff, "/usr/bin/rrdtool graph icmp.png --start %ld --end %ld --title=\"ICMP Activity (%s)\" --vertical-label \"Bytes\" ",
				   tt-GRAPH_LEN, tt, hostname);
	strcat (buff, "DEF:ta=icmp.rrd:icmp_accept:LAST DEF:td=icmp.rrd:icmp_drop:LAST \"CDEF:test=ta,50,+,50,-\" ");
	strcat (buff, "AREA:test#CCDDFF LINE2:ta#3300FF:\"ICMP Accept\" ");
	strcat (buff, "'GPRINT:ta:LAST:\%6.0lf bytes' AREA:td#FF0033:\"ICMP Drop\" 'GPRINT:td:LAST:\%6.0lf bytes\\j' ");
	strcat (buff, "> /dev/null");
	system (buff);

	// ip
	memset (buff, 0, 1024);
	sprintf(buff, "/usr/bin/rrdtool update ip.rrd %ld:%ld:%ld", tt, ACCEPT_BYTE(ip), DROP_BYTE(ip));
	system (buff);
	memset (buff, 0, 1024);
	sprintf(buff, "usr/bin/rrdtool graph ip.png --start %ld --end %ld --title=\"IP Activity (%s)\" --vertical-label \"Bytes\" ",
				   tt-GRAPH_LEN, tt, hostname);
	strcat (buff, "DEF:ta=ip.rrd:ip_accept:LAST DEF:td=ip.rrd:ip_drop:LAST \"CDEF:test=ta,50,+,50,-\" ");
	strcat (buff, "AREA:test#CCDDFF LINE2:ta#3300FF:\"IP Accept\" ");
	strcat (buff, "'GPRINT:ta:LAST:\%6.0lf bytes' AREA:td#FF0033:\"IP Drop\" 'GPRINT:td:LAST:\%6.0lf bytes\\j' ");
	strcat (buff, "> /dev/null");
	system (buff);
}

void update_rule_stats (char *s, char *rule_desc, struct prot_stat *ps)
{
	struct stat statbuf;
	char buff[1024];
	char rrdfile[80];
	unsigned long tt = time (NULL);

	memset (rrdfile, 0, 80);
	sprintf (rrdfile, "%s.rrd", s);

	if (stat(rrdfile, &statbuf)<0) {
		memset (buff, 0, 1024);
		sprintf(buff, "/usr/bin/rrdtool create %s.rrd --start %ld --step %d DS:%s_packet:GAUGE:%d:U:U DS:%s_bytes:GAUGE:%d:U:U RRA:LAST:0.5:1:1000 RRA:AVERAGE:0.5:1:1000", s, tt-10, STEP, s, STEP, s, STEP);
		system (buff);
	}

	memset (buff, 0, 1024);
	sprintf(buff, "/usr/bin/rrdtool update %s.rrd %ld:%ld:%ld", s, tt, ps->count, ps->byte);
	system (buff);

	memset (buff, 0, 1024);
	sprintf(buff, "/usr/bin/rrdtool graph %s.png --start %ld --end %ld --title=\"%s (%s)\" --vertical-label \"Bytes\" DEF:ta=%s.rrd:%s_packet:LAST DEF:td=%s.rrd:%s_bytes:LAST \"CDEF:test=ta,50,+,50,-\" AREA:test#CCDDFF LINE2:ta#3300FF:\"Packets\" AREA:test#FF0033:\"Bytes\" > /dev/null", s, tt-GRAPH_LEN, tt, rule_desc, hostname, s, s, s, s);
	system (buff);
}

char *request_rule (int sock_fd, struct msghdr *mymsg, void *key)
{
	int size = 4;
	char type = OPT_READ_RULE;
	char *rule = NULL;
	int len;
	struct nlmsghdr *nlmh = NULL;
	struct sockaddr_nl src, dst;
	struct iovec myiov;
	int mysockfd;

	if ((mysockfd = socket(PF_NETLINK, SOCK_RAW, NETLINK_GRAPH))<0) {
		perror("socket");
		exit(0);
	}

	src.nl_family = AF_NETLINK;      
	src.nl_pid = getpid();  
	src.nl_groups = 0;  

	if ((bind(sock_fd, (struct sockaddr*)&src, sizeof(src)))<0) {
		perror("bind");
		exit(0);
	}

	dst.nl_family = AF_NETLINK;
	dst.nl_pid = 0;   
	dst.nl_groups = 0; 
	if ((nlmh=(struct nlmsghdr *)malloc(NLMSG_SPACE(MAX_PAYLOAD)))==NULL) {
		perror("malloc");
		exit(0);
	}

	/* Fill the netlink message header */
	nlmh->nlmsg_len = NLMSG_SPACE(MAX_PAYLOAD);
	nlmh->nlmsg_pid = getpid(); 
	nlmh->nlmsg_flags = 0;

	/* Fill in the netlink message payload */
	size = 4;
	memcpy (NLMSG_DATA(nlmh), &type, 1);
	memcpy (NLMSG_DATA(nlmh)+1, &size, 4);
	memcpy (NLMSG_DATA(nlmh)+5, key, 4);
	myiov.iov_base = (void *)nlmh;
	myiov.iov_len = nlmh->nlmsg_len;
	mymsg->msg_name = (void *)&dst;
	mymsg->msg_namelen = sizeof(dst);
	mymsg->msg_iov = &myiov;
	mymsg->msg_iovlen = 1;

	if ((sendmsg(sock_fd, mymsg, 0))<0) {
		perror("sendmsg");
		exit(0);
	}

	memset(nlmh, 0, NLMSG_SPACE(MAX_PAYLOAD));
	if ((len = recvmsg(sock_fd, mymsg, 0))<0) {
		perror("recvmsg");
	}

	rule = (char *) (NLMSG_DATA(nlmh)+5);

	return rule;
}

void request_all_rules (int sock_fd, struct msghdr *mymsg)
{
	int size = 4;
	char type = OPT_READ_ALL_RULES;
	struct nlmsghdr *nlmh = NULL;
	struct sockaddr_nl src, dst;
	struct iovec myiov;

	src.nl_family = AF_NETLINK;      
	src.nl_pid = getpid();  
	src.nl_groups = 0;  

	if ((bind(sock_fd, (struct sockaddr*)&src, sizeof(src)))<0) {
		perror("bind");
		exit(0);
	}

	dst.nl_family = AF_NETLINK;
	dst.nl_pid = 0;   
	dst.nl_groups = 0; 
	if ((nlmh=(struct nlmsghdr *)malloc(NLMSG_SPACE(MAX_PAYLOAD)))==NULL) {
		perror("malloc");
		exit(0);
	}

	/* Fill the netlink message header */
	nlmh->nlmsg_len = NLMSG_SPACE(MAX_PAYLOAD);
	nlmh->nlmsg_pid = getpid(); 
	nlmh->nlmsg_flags = 0;

	/* Fill in the netlink message payload */
	size = 4;
	memcpy (NLMSG_DATA(nlmh), &type, 1);
	myiov.iov_base = (void *)nlmh;
	myiov.iov_len = nlmh->nlmsg_len;
	mymsg->msg_name = (void *)&dst;
	mymsg->msg_namelen = sizeof(dst);
	mymsg->msg_iov = &myiov;
	mymsg->msg_iovlen = 1;

	if ((sendmsg(sock_fd, mymsg, 0))<0) {
		perror("sendmsg");
		exit(0);
	}

}

void create_datafile (char *dir)
{
	int i;
	char file[80];
	FILE *fp = NULL;

	memset (file, 0, 80);
	sprintf(file, "%s/graph.ini", dir);

	if ((fp = fopen(file, "w+"))==NULL) {
		perror("fopen");
		exit(0);
	}

	for (i=0; i<last_rstats_idx; i++) {
		memset(file, 0, 80);
		sprintf(file, "rule%01d", i);
		fprintf (fp, "%s %-50s\n", file, rule_stats[i].rule_desc);
	}
	fclose (fp);
}

void remove_files (char *dir)
{
	DIR *dp = NULL;
	struct dirent *dirp;
	char file[1024];

	if (!(dp = opendir(dir))) {
		perror("opendir");
		exit(0);
	}

	while ((dirp = readdir(dp))) {
		if (!strcmp(dirp->d_name, ".") || !strcmp(dirp->d_name, ".."))
			continue;

		if (!strstr(dirp->d_name, "rule"))
			continue;

		if (!strstr(dirp->d_name, "rrd") && !strstr(dirp->d_name, "png"))
			continue;

		memset(file, 0, 1024);
		sprintf(file, "%s/%s", dir, dirp->d_name);

		unlink(file);
	}

	closedir(dp);	
	
}

void daemonize (void)
{
	int i, fd0, fd1, fd2;
	pid_t pid;
	struct rlimit r1;
	struct sigaction sa;

	umask(0);
	if (getrlimit(RLIMIT_NOFILE, &r1)<0) {
		perror("getrlimit");
		exit(0);
	}

	if ((pid=fork())<0) {
		perror("fork");
		exit(0);
	} else if (pid!=0) {
		exit(0);
	}
	setsid();

	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if (sigaction(SIGHUP, &sa, NULL) < 0) {
		perror("sigaction");
		exit(0);
	} 
	if ((pid=fork())<0) {
		perror("fork");
		exit(0);
	} else if (pid!=0) {
		exit(0);
	}

	
	if (chdir(destdir)<0) {
		perror("chdir");
	}
	
	if (r1.rlim_max == RLIM_INFINITY)
		r1.rlim_max = 1024;
	for (i=3; i<r1.rlim_max; i++)
		close(i);
	
}

int monitor (void)
{
	int grp = 1;
	struct total_stats *tstats = NULL;
	struct prot_stat *ps = NULL;
	char *type;
	int *length;
	char filename[20];
	void *key;
	int recvlen;
	int i, j;
	int found;
	int write_file = 1;

	hostname = myhostname();
	create_rrd();
	if ((sock_fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_GRAPH))<0) {
		perror("socket");
		exit(0);
	}

	i = j = 0;
	last_rstats_idx = 0;
	found = 0;

	request_all_rules (sock_fd, &msg);
	while(1) {
		memset (&src_addr, 0, sizeof(src_addr));
		memset (&msg, 0, sizeof(msg));
		src_addr.nl_family = AF_NETLINK;
		src_addr.nl_pid = getpid();
		src_addr.nl_groups = grp;
		if ((bind (sock_fd, (struct sockaddr *)&src_addr, sizeof(src_addr)))<0) {
			perror("bind");
			exit(0);
		}

		nlh = (struct nlmsghdr *) malloc (NLMSG_SPACE(MAX_PAYLOAD));
		memset (nlh, 0, NLMSG_SPACE(MAX_PAYLOAD));

		memset (&dest_addr, 0, sizeof(dest_addr));
		dest_addr.nl_family = AF_NETLINK;
		dest_addr.nl_pid = 0;
		dest_addr.nl_groups = 0;
		setsockopt (sock_fd, SOL_SOCKET, SO_RCVBUF, nlh, NLMSG_SPACE(MAX_PAYLOAD));

		iov.iov_base = (void *)nlh;
		iov.iov_len = NLMSG_SPACE(MAX_PAYLOAD);
		msg.msg_name = (void *)&dest_addr;
		msg.msg_namelen = sizeof(dest_addr);
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;

		if ((recvlen = recvmsg (sock_fd, &msg, 0))<0) {
			perror("recvmsg");
			exit(0);
		}


		type = (char *) NLMSG_DATA(nlh);

		switch (*type) {
			case OPT_PROTO_STATS:
				length = (int *) (NLMSG_DATA(nlh)+1);		
				tstats = (struct total_stats *) (NLMSG_DATA(nlh)+5);

				if (verbose)
					print_stats (tstats);
				update_rrd_stats (tstats);
				i = 0;

			break;

			case OPT_RULE_STATS:
				length = (int *) (NLMSG_DATA(nlh)+ 1);		
				key = (void *) (NLMSG_DATA(nlh) + 5);
				ps = (struct prot_stat *) (NLMSG_DATA(nlh) + 9);
				found = 0;

				// check to see if key already exists
				while (i<last_rstats_idx+1) {
					if (rule_stats[i].key != NULL && !memcmp(rule_stats[i].key, key, 4)) {
						break;
					} else if (rule_stats[i].key == NULL) {
						rule_stats[i].key = key;
						last_rstats_idx++;
						found = 1;
						break;
					}
					i++;
				}

				rule_stats[i].ps.count = ps->count;
				rule_stats[i].ps.byte = ps->byte;

			break;

			case OPT_FLUSHTBL:
				remove_files (destdir);
				write_file = 1;
				request_all_rules (sock_fd, &msg);
			break;

			case OPT_LAST_IDX:
				for (i=0; i<last_rstats_idx; i++) {
					sprintf(filename, "rule%01d", i);
					if (rule_stats[i].rule_desc == NULL) {
							rule_stats[i].rule_desc =
							request_rule (sock_fd, &msg, rule_stats[i].key);
					}

					if (verbose)
						printf ("%s %-50s count %-6ld byte %-6ld\n", 
							filename, rule_stats[i].rule_desc, 
							rule_stats[i].ps.count, rule_stats[i].ps.byte);

					update_rule_stats (filename, 
									rule_stats[i].rule_desc,
									&rule_stats[i].ps);

					rule_stats[i].ps.count = 0;
					rule_stats[i].ps.byte = 0;

				}

				if (write_file) {
					create_datafile (destdir);
					write_file = 0;
				}

			break;
		}

	}
	close(sock_fd);

	return 0;
}

void help (void)
{
	printf("recv -d <directory> | -v | -h | -V\n");
	printf("options:\n");
	printf("-d <directory>		: dest directory\n");
	printf("-v			: verbose\n");
	printf("-V			: version\n");
	printf("-h			: help\n");
	exit(0);
}

int verify_directory (char *dir)
{
	struct stat buf;

	if (dir == NULL)
		help();

	if (lstat(dir, &buf)<0) {
		perror("lstat");
		exit(0);
	}

	if (!S_ISDIR(buf.st_mode)) 
		return 0;
	else 
		return 1;
}

void check_rrdtool (void)
{
	struct stat buf;

	if (lstat("/usr/bin/rrdtool", &buf)<0) {
		perror("/usr/bin/rrdtool not found! Please installed rrdtool first.");
		exit(0);
	}

}

void print_version (void)
{
	printf("%s\n", version);
	exit(0);
}

int main (int argc, char **argv)
{
	int c;
	extern char *optarg;
	destdir = NULL;
	char *tmp = NULL;
	int len = 0;

	while ((c = getopt(argc, argv, "d:vhV")) != EOF) {
		switch (c) {
			case 'd':
				destdir = optarg;
				len = strlen(destdir)-1;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'V':
				print_version();
				break;
			case 'h':
			default :
				help();
				break;
		}
	}

	if (!verify_directory(destdir)) 
		help();

	/* remove trailing slashes */
	while (*(destdir+len) == '/') {
		*(destdir+len) = '\0';
		len--;
	}

	check_rrdtool();
	
	daemonize();

	monitor();

	return 0;
}

